/// CaveGame - A procedural 2D platformer sandbox.
/// Created by Josh O'Leary @ Redacted Software, 2020-2025
/// Contact: josh@redacted.cc
/// Contributors: william@redacted.cc maxi@redacted.cc
/// This work is dedicated to the public domain.

/// @file GameSession.hpp
/// @desc The scene that runs and manages a client game session.
/// @edit 1/28/2025
/// @auth Josh O'Leary

#pragma once

#include <Client/Scene.hpp>
#include <J3ML/LinearAlgebra.hpp>
#include <Core/Entity.hpp>
#include <Core/Player.hpp>
#include <Core/World.hpp>
#include "LocalWorld.hpp"
#include "Hotbar.hpp"
#include <JUI/Widgets/Scene.hpp>
#include <JUI/Widgets/Rect.hpp>
#include <JUI/Widgets/TextRect.hpp>
#include <Core/Item.hpp>
#include "TileTool.hpp"
#include "ContainerWindow.hpp"
#include <Client/PauseMenu.hpp>

namespace CaveGame::Client {
    using namespace Core;

    class LootTable {};
    class CraftingStation {};
    class ContainerGUI {};

    using namespace JUI::UDimLiterals;
    using CaveGame::Core::Entity;
    class GameSession : public Scene {
    public:

        Event<> OnSessionExit;
        Event<> RequestToggleSettings;

        GameSession() = default;
        explicit GameSession(bool createNewWorld);
        void Update(float elapsed) override;
        void Draw() override;
        void Load() override;
        void Unload() override;
        void PassWindowSize(const J3ML::LinearAlgebra::Vector2 &size) override;
        void PassMouseInput(unsigned int a, bool b) override;

        void PassMouseMovement(const J3ML::LinearAlgebra::Vector2 &pos) override;

        void PassMouseWheel(int wheel) override;

        void PassKeyInput(Key key, bool pressed) override;

        void SaveAndExit();

        void UnloadHUD();

        uint16_t GetTileIDUnderMouse();

        [[nodiscard]] JUI::Scene* HUD() const { return hud; }
        LocalWorld* World() const { return world;}
        TileHotbar Hotbar() const { return hotbar;}

        void WorldEditToolControlsUpdate(float elapsed);
        void ToggleWorldEdit();
        TileTool* WorldEditToolWindow() const { return tile_tool;}

        Core::Player* GetLocalPlayerEntity();
    protected:
        TileTool* tile_tool = nullptr;
        LocalWorld* world = nullptr;
        TileHotbar hotbar;
        JUI::Scene* hud = nullptr;
        PauseMenuWidget* pause_menu = nullptr;
        Vector2 mouse_pos = {0,0};
        RNG tool_rng;

    private:
        void WorldEditToolDrawTiles(TileID tile);
        void WorldEditToolDrawTiles(int x, int y, int radius, int density, TileID tile);
        void WorldEditToolDrawOverlay();
        Vector2 MouseWorldPos() const;
    };
}
