/// CaveGame - A procedural 2D platformer sandbox.
/// Created by Josh O'Leary @ Redacted Software, 2020-2025
/// Contact: josh@redacted.cc
/// Contributors: william@redacted.cc maxi@redacted.cc
/// This work is dedicated to the public domain.

/// @file LocalWorld.hpp
/// @desc The client implementation for game world behavior.
/// @edit 1/28/2025
/// @auth Josh O'Leary

#pragma once

#include <Core/World.hpp>
#include <Core/Chunk.hpp>
#include <JGL/JGL.h>
#include <Client/Camera2D.hpp>
#include <Core/SimpleAABBSolver.hpp>
#include <thread>
#include <Core/ConcurrentQueue.hpp>
#include <Core/ThreadPool.hpp>
#include "Particle.hpp"

namespace CaveGame::Client
{
    using namespace std::chrono_literals;

    class AsyncChunkRenderer : public ThreadPool
    {
    public:

    };


    /// A class object that renders an overlay on chunks for debugging. This includes grid-lines, cell coordinates, activity stats, etc.
    class ChunkDebugGizmo
    {
    public:

        void DrawGrid(const AABB2D& bounds) {
            Vector2 viewport_topleft = bounds.minPoint;
            Vector2 viewport_bottomright = bounds.maxPoint;

            int nearest_grid_left = Math::Floor(viewport_topleft.x / Core::Chunk::ChunkSize);
            int nearest_grid_right = Math::Floor(viewport_bottomright.x / Core::Chunk::ChunkSize);

            for (int x = nearest_grid_left; x <= nearest_grid_right; x++) {
                auto top = Vector2(x * Core::Chunk::ChunkSize, viewport_topleft.y);
                auto bottom = Vector2(x * Core::Chunk::ChunkSize, viewport_bottomright.y);
                JGL::J2D::DrawLine(grid_color, top, bottom, 1 / bounds.Width());
            }

            int nearest_grid_top = Math::Floor(viewport_topleft.y / Core::Chunk::ChunkSize);
            int nearest_grid_bottom = Math::Floor(viewport_bottomright.y / Core::Chunk::ChunkSize);

            for (int y = nearest_grid_top; y <= nearest_grid_bottom; y++) {
                auto left = Vector2(viewport_topleft.x, y * Core::Chunk::ChunkSize);
                auto right = Vector2(viewport_bottomright.x, y * Core::Chunk::ChunkSize);
                JGL::J2D::DrawLine(grid_color, left, right, 1 / bounds.Height());
            }
        }
        void DrawCellCoords(const Vector2i& coords, float scale) {
            // TODO: Fix offset on text.
            JGL::J2D::DrawString(Colors::Black, std::format("{}, {}", coords.x, coords.y), coords.x * Core::Chunk::ChunkSize, (coords.y * Core::Chunk::ChunkSize)-5, scale, 10);
        }
        void DrawStats(Core::Chunk* chunk, float scale) {
            Vector2i cell = chunk->GetChunkCell();
            Vector2i coords = chunk->GetChunkRealCoordinates();

            JGL::J2D::DrawString(Colors::Black, std::format("RenderTarget Age: {}", chunk->time_since_refresh), coords.x, coords.y, scale, 10);
        }
        void DrawHeatmap() {}

        void Enable(bool enable = true)
        {
            this->enabled = enable;
        }
        void Disable() { Enable(false); }
        [[nodiscard]] bool IsEnabled() const { return enabled; }

    protected:
        bool enabled = false;
        Color4 grid_color = {128, 128, 128, 128};

    private:
    };

    /// An extension of the world object that provides rendering for the game world.
    class LocalWorld : public CaveGame::Core::World {

    public:
        LocalWorld(const std::string& world_name, int seed, bool overwrite);
        void PostInit();
        void Draw();
        void Update(float elapsed) override;
        void RefreshAll() override;
        void DebugChunks(bool enabled);

        unsigned int GetRenderTargetCount() const;

        void SaveAndExit() override;

        void SetShowTileActivity(bool enabled);

        bool IsShowTileActivityEnabled() const;

        void RenderTile(const Core::TileID &tid, int wx, int wy, int tx, int ty);

        //Camera2D* Camera()  { return &camera;}

        void Emit(const Particle& p);

    protected:
        /// Determines whether a given chunk coordinate lies within the viewable space, with an optional 'oversize' parameter.
        /// @param extraChunkRadius Specifies how many extra chunks can be considered within the viewport when they are in fact, just outside of the viewport.
        /// @note This is used to overdraw
        bool IsChunkCellWithinViewport(const Vector2i& coords, int extraChunkRadius = 0) const;
        void RenderChunk(const Vector2i& coords);
        void LookForChunksNeedLoading();
        void LookForChunksNeedUnloading();
        void RenderChunkTexture(const Vector2i &coords, JGL::RenderTarget* destination, Core::Chunk* chunk);
        /// Checks for missing, or out-of-date, cached sprites of chunks, and renders them.
        void CheckCachedChunkSprites();
        static Color4 GetSkyColorInterpolatedForTimeOfDay(float time);
        static Color4 GetSkyColorBaseForTimeOfDay(float time);
        void DrawSky();

    public:
        Camera2D camera;
        Vector2 mouse_pos;
        ChunkDebugGizmo chunk_debug;
        JGL::Font font;
    protected:

        float check_chunks_timer = 0.f;

        std::unordered_map<Vector2i, JGL::RenderTarget*> cached_chunk_sprites;
        JGL::RenderTarget* canvas_render_target;

        std::vector<Particle> particles;



        bool show_tile_activity = false;

    };
}