/// CaveGame - A procedural 2D platformer sandbox.
/// Created by Josh O'Leary @ Redacted Software, 2020-2024
/// Contact: josh@redacted.cc
/// Contributors: william@redacted.cc maxi@redacted.cc
/// This work is dedicated to the public domain.

/// @file CaveGameWindow.hpp
/// @desc The Client App Window. See ReWindow's Documentation.
/// @edit 11/1/2024
/// @auth Josh O'Leary

#pragma once

#include <Client/AssetService.hpp>

#include <J3ML/Geometry.hpp>
#include <Client/Splash.hpp>
#include <Client/MainMenu.hpp>
#include <Client/GameSession.hpp>
#include <JUI/Widgets/ListLayout.hpp>
#include <JUI/UDim.hpp>
#include <ReWindow/types/Window.h>
#include <JUI/Widgets/Window.hpp>
#include <Client/Scene.hpp>
#include <Client/SceneManager.hpp>
#include <JUI/Widgets/CommandLine.hpp>
#include <Client/TileTool.hpp>
#include "Command.hpp"
#include <Core/Player.hpp>
#include <Core/Explosion.hpp>


namespace CaveGame::ClientApp {

    using CaveGame::Client::Scene;
    using CaveGame::Client::SceneManager;

    using CommandArgs = std::vector<std::string>;

    /// The main program class. Everything originates here.
    /// This class is derived from RWindow class.
    class CaveGameWindow : public ReWindow::OpenGLWindow, public SceneManager {
    public:
        /// Constructs and returns an instance of the game program.
        /// @param title
        /// @param width
        /// @param height
        CaveGameWindow(const std::string& title, int width, int height);
        /// Destructor called when deleting the CaveGameWindow.
        ~CaveGameWindow() override;

        /// This function runs the totality of the game loop, start to finish.
        /// Calls Open, Init, Gameloop(), and Cleanup() in that order
        /// @note The game loop runs until Die() is called, at which point final cleanup is performed.
        void Run();
        /// This function triggers an internal signal that the game window is ready to close.
        /// @note This does not immediately close the window, rather, time is given to allow for final data saving and other processes.
        void Die(); void MarkReadyToClose(bool close = true);
        /// This function sets up the initial program state, particularly that which must be performed **after** the window is opened.
        void Init();
        /// This function cleans up any data or assets before closing the game.
        void Cleanup();
        /// This function performs logic calculation routines, once per refresh.
        void Update(float elapsed);
        /// This function performs rendering routines, once per refresh.
        void Draw();

        /// @return True when a flag is set that indicates the game is "preparing" to close. Certain procedures may still be taking place, such as saving and serialization.
        bool ReadyToClose() const;

        /// Returns the game's console GUI, which accepts commands and displays log messages.
        JUI::CommandLine* Console();
        /// Returns the Scene that holds global menu windows, which can appear in any game-context.
        JUI::Scene* WindowManager();
        /// Returns the active game-world session, if it exists.
        Client::GameSession* GameSession();

        /// @return the user's mouse coordinates from the RWindow layer.
        [[nodiscard]] Vector2 GetMouseV2() const;
        /// @return the window's size from the RWindow layer.
        [[nodiscard]] Vector2 GetSizeV2() const;
        /// @return true if the game is currently in-session.
        [[nodiscard]] bool InGame() const;
        /// @returns true if the game is currently in the main menu.
        [[nodiscard]] bool InMainMenu() const;

    public:
#pragma region Input Callbacks
        /// Called by the window on each frame.
        void OnRefresh(float elapsed) override;
        /// Called by the window upon the user pressing a mouse button.
        void OnMouseButtonDown(const ReWindow::MouseButtonDownEvent &ev) override;
        /// Called by the window upon the user releasing a mouse button.
        void OnMouseButtonUp(const ReWindow::MouseButtonUpEvent &ev) override;
        /// Called by the window upon the user scrolling a mouse wheel.
        void OnMouseWheel(const ReWindow::MouseWheelEvent &ev) override;
        /// Called by the window upon the user releasing a keyboard key.
        void OnKeyUp(const ReWindow::KeyUpEvent &ev) override;
        /// Called by the window upon the user pressing a keyboard key.
        void OnKeyDown(const ReWindow::KeyDownEvent& ev) override;
        /// Called by the window upon the user moving their pointer device, currently just mice.
        void OnMouseMove(const ReWindow::MouseMoveEvent &ev) override;
        /// Called by the window when it receives a request from the operating-system to resize.
        bool OnResizeRequest(const ReWindow::WindowResizeRequestEvent &ev) override;
        /// Called by the window **before** it closes.
        void OnClosing() override;
#pragma endregion
    public:
        /// Forwards a message to the console log.
        void Log(const std::string& msg, const Color4& color = Colors::White);
        bool HasCommand(const std::string &command);
        Command GetCommand(const std::string &command);
    protected:

        /// Calls Step() in a loop until the window wants to close.
        void Gameloop();
        /// Runs exactly one iteration of the game loop. Currently, this is one call to Update(), and then to Draw().
        /// @see Refresh().
        void Step();
        /// Creates the in-game console menu.
        void create_console_window();
        void create_stats_window();
        void create_settings_window();
        void create_window_widgets();
        /// Draws a sequence of 'debug' information strings to the screen, in a descending list.
        void draw_debug_info(std::vector<std::string> tokens);
        void CreateMenuWindows();
        /// Constructs the Splash screen, Main Menu screen, and In-game session.
        void CreateContexts();



        // TODO: Refactor this into irrelevance.
        void InGameControls(float elapsed);
        void OnConsoleCommandInput(const std::string &command);
        /// Toggles tile simulation if we are in-game.
        bool ToggleTileSim(const CommandArgs& args);
        /// Opens a singleplayer session.
        /// @param info A structure containing information for joining a singleplayer world.
        void OpenWorld(Client::SingleplayerSessionInfo info);

        /// Logs help information to the console. Called when the user runs the 'help' command.
        bool HelpCommand(const CommandArgs& args);
        /// Shows a list of every command label.
        bool ListCommand(const CommandArgs& args);
        bool TileListCmd(const CommandArgs& args);
        bool ItemListCmd(const CommandArgs& args);


        bool NoclipCmd(const CommandArgs &args);

        bool FpsLimitCmd(const CommandArgs &args);

        /// This table defines the supported console commands, their aliases, and the callback lambda.
#pragma region Commands
        const std::vector<Command> commands {
{"help", {}, [this](const CommandArgs& args) { HelpCommand(args); }},
{"list", {}, [this](const CommandArgs& args) {  ListCommand(args); }},
{"quit", {"q"}, [this](const CommandArgs& args) { Close(); }},
{"worldedit", {"we"}, [this](const CommandArgs& args) {
    if (InGame())
        GameSession()->ToggleWorldEdit();
    //tile_tool->Enable(!tile_tool->IsEnabled());
    return true;
}},
{"tilesim", {"q"}, [this](const CommandArgs& args) { return ToggleTileSim(args);}},
{"grid", {"q"}, [this](const CommandArgs& args) {}},
{"credits", {"q"}, [this](const CommandArgs& args) { credits_window->Toggle(); }},
{"settings", {"q"}, [this](const CommandArgs& args) {}},
{"spawn", {}, [this](const CommandArgs& args) {
    if (!InGame())
        return Log("Error: This command is only available when in a world!", Colors::Red);

    if (args.size() == 1)
        return Log("Available entities: player, explosion, particle");

    if (args[1] == "player")
    {
        auto pos = GameSession()->World()->camera.ScreenToWorld(GetMouseV2());
        auto* plr = new CaveGame::Core::Player(pos);
        GameSession()->World()->AddEntity(plr);
        return;
    }
    if (args[1] == "explosion")
    {
        auto pos = GameSession()->World()->camera.ScreenToWorld(GetMouseV2());
        auto* plr = new CaveGame::Core::Explosion(GameSession()->World(), pos, 1.f, GameSession()->WorldEditToolWindow()->BrushRadius());
        GameSession()->World()->AddEntity(plr);
        return;
    }
    if (args[1] == "particle")
    {
        /*auto pos = GameSession()->World()->camera.ScreenToWorld(GetMouseV2());
        auto* plr = new CaveGame::Core::Player(pos);
        GameSession()->World()->AddEntity(plr);
        return;*/
    }
}},
{"settings", {"q"}, [this](const CommandArgs& args) {}},
{"give", {"q"}, [this](const CommandArgs& args) {}},
{"tp", {"q"}, [this](const CommandArgs& args) {}},
{"freecam", {"q"}, [this](const CommandArgs& args) {}},
{"god", {"q"}, [this](const CommandArgs& args) {}},
{"fullbright", {"q"}, [this](const CommandArgs& args) {}},
{"connect", {"q"}, [this](const CommandArgs& args) {}},
{"disconnect", {"q"}, [this](const CommandArgs& args) {}},
{"server", {"q"}, [this](const CommandArgs& args) {}},
{"sandbox", {"q"}, [this](const CommandArgs& args) {}},
{"crash", {"q"}, [this](const CommandArgs& args) {}},
{"time", {"q"}, [this](const CommandArgs& args) {}},
{"juidebug", {"q"}, [this](const CommandArgs& args) {}},
{"fpslimit", {"fps"}, [this](const CommandArgs& args) { return FpsLimitCmd(args); }},
{"grid", {}, [this](const CommandArgs& args) {
    if (current_scene == game_ctx)
    {
        game_ctx->World()->chunk_debug.Enable(!game_ctx->World()->chunk_debug.IsEnabled());
    } else
        Log("Must be in-game to use this command!");
}},
{"tileactivity", {}, [this](const CommandArgs& args) {
    if (current_scene == game_ctx)
    {
        game_ctx->World()->SetShowTileActivity(!game_ctx->World()->IsShowTileActivityEnabled());
    } else
        Log("Must be in-game to use this command!");
}},
{"save-world", {}, [this](const CommandArgs& args) {
   if (InGame())
       GameSession()->World()->Save();
}},
{"tile-list", {}, [this](const CommandArgs& args) { return TileListCmd(args); }},
{"item-list", {}, [this](const CommandArgs& args) { return ItemListCmd(args); }},
{"container", {}, [this](const CommandArgs& args) {
    Log(std::format("Size: {}", args.size()));
    if (args.size() == 3)
    {
        int w = std::stoi(args[1]);
        int h = std::stoi(args[2]);

        GameSession()->HUD()->Add(new ContainerWindow(GameSession()->HUD(), w, h, "Test Container"));
    }
}},
{"noclip", {}, [this](const CommandArgs& args) { return NoclipCmd(args); }}
};

        Command InvalidCommand {
                "n/a", {}, [this](const CommandArgs& args) {
                    Log("Nope");
                }
        };
#pragma endregion



    protected: // Complex class members.
        Client::AssetService assets;
        CaveGame::Client::Splash* splash_ctx = nullptr;
        CaveGame::Client::MainMenu* menu_ctx = nullptr;
        CaveGame::Client::GameSession* game_ctx = nullptr;
        JUI::Scene* wm = nullptr;
        JUI::Window* settings_window = nullptr;
        JUI::Window* credits_window = nullptr;

        JUI::CommandLine* console_window = nullptr;
        JUI::Window* stats_window = nullptr;
    protected: // Primitive class members.
        /// This field indicates the program is ready to close.
        bool wanna_die = false;
        bool render_grid = true;
        bool generate_grid = true;
        bool mbd = false;
        float our_avg = 0.f;
        float max_fps = 60.f;


    };
}