#include <Core/Chunk.hpp>
#include "Core/Perlin.hpp"
#include "Core/Simplex.hpp"
#include "Core/Data.hpp"
#include <iostream>
#include <Core/TileRegistry.hpp>

namespace CaveGame::Core
{
    void Chunk::SetTile(int x, int y, TileID t, bool trigger_tile_updates) {
        if (!(0 <= x && x < ChunkSize && 0 <= y && y < ChunkSize))
            assert("Out of bounds!");

        if (tiles[x][y] != t) {
            const Tile& old_tile = Tiles().GetByNumericID(tiles[x][y]);
            const Tile& new_tile = Tiles().GetByNumericID(t);

            if (old_tile.does_random_ticc)
                tiles_with_random_tick_count--;
            if (new_tile.does_random_ticc)
                tiles_with_random_tick_count++;

            tiles[x][y] = t;
            touched = true;
            if (trigger_tile_updates)
                SetTileUpdateFlag(x, y);
        }
    }

    void Chunk::SetTileSilent(int x, int y, TileID t)
    { SetTile(x, y, t, false); }

    bool Chunk::GetTileUpdateFlag(int x, int y) const
    {
        size_t index = y * ChunkSize + x;
        return tagged_for_update.test(index);
    }

    bool Chunk::GetTileUpdateBufferFlag(int x, int y) const {
        size_t index = y * ChunkSize + x;
        return update_buffer.test(index);
    }

    void Chunk::SetTileUpdateBufferFlag(int x, int y, bool flag) {
        size_t index = y * ChunkSize + x;

        if (flag)
            update_buffer.set(index);
        else
            update_buffer.reset(index);
    }

    void Chunk::SetTileUpdateFlag(int x, int y, bool flag)
    {
        size_t index = y * ChunkSize + x;

        if (flag)
            tagged_for_update.set(index),
            has_update = true;
        else
            tagged_for_update.reset(index);
    }

    Chunk::Chunk(const Vector2i& cell_coords) : cell(cell_coords), touched(false), update_buffer{}, tagged_for_update{}, tiles{} {}

    Chunk::Chunk(const Vector2i& cell_coords, const std::filesystem::path& file): Chunk(cell_coords)
    {
        std::ifstream inp;

        inp.open(file, std::ios::binary | std::ios::in);
        inp.seekg(0, std::ios::end);
        int length = inp.tellg();
        inp.seekg(0, std::ios::beg);

        char* buffer = new char[length];
        inp.read(buffer, length);
        inp.close();

        SetData(buffer);

        delete buffer;
    }

    Vector2i Chunk::GetChunkRealCoordinates() const {
        return GetChunkCell()*Core::Chunk::ChunkSize;
    }

    std::array<TileID, Chunk::ChunkSize * Chunk::ChunkSize> Chunk::DataContiguous() const {
        std::array<TileID, ChunkSize * ChunkSize> result{};
        memcpy(result.data(), tiles, ChunkSize * ChunkSize * sizeof(TileID));
        return result;
    }

    void Chunk::SetData(char* buffer) {
        auto* data = reinterpret_cast<TileID*>(buffer);
        memcpy(tiles, data, ChunkSize * ChunkSize * sizeof(TileID));
    }

    void Chunk::SwapTileUpdateBuffers() {
        memcpy(&update_buffer, &tagged_for_update, sizeof(tagged_for_update));
        tagged_for_update.reset();
        has_update = false;
    }

    bool Chunk::FirstPassCompleted() const { return first_pass_complete;}

    bool Chunk::SecondPassCompleted() const { return second_pass_complete;}

    void Chunk::SetFirstPassCompleted(bool complete) { first_pass_complete = complete; }

    void Chunk::SetSecondPassCompleted(bool complete) { second_pass_complete = complete; }

    Vector2i Chunk::GetChunkCell() const { return cell;}

}
