#include <Core/PhysicsEntity.hpp>
#include <J3ML/LinearAlgebra/Vector2.hpp>

namespace CaveGame::Core
{
    PhysicsEntity::PhysicsEntity(const Vector2 &spawnPoint): Entity(spawnPoint) {
        next_position = spawnPoint;
    }

    Vector2 PhysicsEntity::Velocity() const { return velocity; }

    Vector2 PhysicsEntity::EstimatedNextPosition() const { return next_position; }

    void PhysicsEntity::Accelerate(const Vector2 &vector) {
        velocity += vector;
    }

    void PhysicsEntity::ApplyHorizontalFriction(float elapsed)
    {
        float base_friction_coefficient = 4.f;

        velocity.x *= (1 - (elapsed * base_friction_coefficient));
    }

    void PhysicsEntity::ApplyGravityForce(float elapsed)
    {
        velocity.y += (elapsed*gravity*Mass());
    }

    void PhysicsEntity::Update(float elapsed) {

        if (!on_ground)
            airtime += elapsed;
        else
            airtime = 0;

        // TODO: Sophisticated mechanism to maintain locked-timestep, multiple-iteration physics steps.
        //PhysicsUpdate(elapsed);

    }

    void PhysicsEntity::ApplyAirResistance(float coefficient, float elapsed)
    {
        velocity *= (1 - (elapsed * coefficient));
    }

    void PhysicsEntity::PhysicsUpdate(float elapsed) {

        if (noclip)
        {
            on_ground = true;

            next_position += velocity * elapsed;

            ApplyAirResistance(air_resistance*4, elapsed);

            position = next_position;

            return;
        }

        next_position += velocity * elapsed;

        if (Math::Abs(velocity.y) > 2)
            on_ground = false;

        // Simulate friction.
        if (on_ground)
            ApplyHorizontalFriction(elapsed);

        ApplyAirResistance(air_resistance, elapsed);
        ApplyGravityForce(elapsed);

        position = next_position;
    }

    void PhysicsEntity::CollisionTest(ITileMap *map, float elapsed) {

        if (noclip)
            return;

        coll_tests = 0;
        coll_hits = 0;

        for (int x = -1; x <= bounding_box.x; x++) {
            for (int y = -1; y <= bounding_box.y; y++) {

                Vector2 coll_topleft = next_position - (bounding_box / 2.f);

                Vector2 tileBoxPos = Vector2(Math::Floor(coll_topleft.x), Math::Floor(coll_topleft.y)) + Vector2(x,y);

                TileID id = map->GetTile(tileBoxPos.x, tileBoxPos.y);

                if (id == Tiles()["void"].numeric_id) {
                    if (freeze_in_void)
                    {
                        velocity = {0,0};
                        next_position = position;
                        return;
                    }
                    continue;
                }


                if (id == Tiles()["air"].numeric_id)
                    continue;

                // TODO: Profile this bit.
                if (Tiles().GetByNumericID(id).solid == false)
                    continue;

                coll_tests++;

                Vector2 entity_centroid = next_position; // + (bounding_box / 2.f);
                Vector2 entity_halfbox = bounding_box / 2.f;

                Vector2 tile_centroid = tileBoxPos + Vector2{0.5f, 0.5f};
                Vector2 tile_halfbox = Vector2{0.5f, 0.5f};

                if (Solver::AABB2Dvs(entity_centroid, entity_halfbox, tile_centroid, tile_halfbox)) {



                    coll_hits++;
                    Vector2 separation = Solver::SolveAABB(entity_centroid, entity_halfbox, tile_centroid, tile_halfbox);

                    Vector2 normal = Solver::GetNormalForAABB(separation, velocity);



                    //std::cout << std::format("Solved: Sep: {},{}, Norm: {},{}", separation.x, separation.y, normal.x, normal.y) << std::endl;

                    if (normal.x == 0 && normal.y == 0)
                        continue;

                    // Touched top.
                    if (normal.y == -1)
                    {
                        velocity.y = 0;
                        on_ground = true;
                        //std::cout << "Touched your feet" << std::endl;
                    }


                    if (normal.y == 1) {
                        velocity.y = -velocity.y*0.5f;
                        //std::cout << "Touched your head" << std::endl;
                    }

                    if (normal.x == -1)
                    {
                        //std::cout << "Touched your right" << std::endl;
                    }

                    if (normal.x == 1)
                    {
                        //std::cout << "Touched your left" << std::endl;
                    }



                    // TODO: Refine condition for "step-up"s
                    if (normal.x != 0 && Math::Abs(velocity.x) > 10 && on_ground) {// && Math::Abs(velocity.y) < 2 ) {
                        //std::cout << "Step up" << std::endl;
                        next_position.y -= 1;
                        //next_position.x -= normal.x;
                        velocity.y -= climbing_skill;
                        velocity.x *= 0.75f;
                    }

                    next_position += separation;


                    //next_position += separation;
                    //return;
                }
            }
        }
    }


}
