#include <Core/Serialization.hpp>
#include <cstdint>
#include <cassert>

#include <J3ML/Algorithm/Reinterpret.hpp>
#include <algorithm>



namespace CaveGame::Core
{

    /* hton16
    ntoh16
    hton32
    ntoh32
    hton64
    ntoh64 */

    template <typename T>
    constexpr T host_to_network(T value) noexcept
    {
#if __BYTE_ORDER == __LITTLE_ENDIAN
        char* ptr = reinterpret_cast<char*>(&value);
        std::reverse(ptr, ptr + sizeof(T));
#endif
        return value;
    }

    template <typename T>
    constexpr T network_to_host(T value) noexcept
    {
        return host_to_network(value);
    }


    void write_u8(Buffer &buffer, uint8_t value) {
        assert(buffer.index + 1 <= buffer.size);
        *(buffer.data + buffer.index) = (value);
        //*((uint8_t*)(buffer.data + buffer.index)) = (value);
        buffer.index += 1;
    }

    void write_u16(Buffer &buffer, uint16_t value) {
        assert(buffer.index + 2 <= buffer.size);
        *((uint16_t*)(buffer.data + buffer.index)) = host_to_network(value);
        buffer.index += 2;
    }

    void write_u32(Buffer &buffer, uint32_t value) {
        assert(buffer.index + 4 <= buffer.size);
        *((uint32_t*)(buffer.data + buffer.index)) = host_to_network(value);
        buffer.index += 4;
    }

    void write_u64(Buffer &buffer, uint64_t value) {
        assert(buffer.index + 8 <= buffer.size);
        *((uint64_t*)(buffer.data + buffer.index)) = host_to_network(value);
        buffer.index += 8;
    }

    uint8_t read_u8(Buffer &buffer) {
        assert(buffer.index + 1 <= buffer.size);
        uint8_t value;
        value = (*((buffer.data + buffer.index)));
        buffer.index += 1;
        return value;
    }

    uint16_t read_u16(Buffer &buffer) {
        assert(buffer.index + 2 <= buffer.size);
        uint16_t value;
        value = network_to_host(*((uint16_t*)(buffer.data + buffer.index)));
        buffer.index += 2;
        return value;
    }

    uint32_t read_u32(Buffer &buffer) {
        assert(buffer.index + 4 <= buffer.size);
        uint32_t value;
        value = network_to_host(*((uint32_t*)(buffer.data + buffer.index)));
        buffer.index += 4;
        return value;
    }



    uint64_t read_u64(Buffer &buffer) {
        assert(buffer.index + 8 <= buffer.size);
        uint64_t value;
        value = network_to_host(*((uint64_t*)(buffer.data + buffer.index)));
        buffer.index += 8;
        return value;
    }


    void write_s8(Buffer& buffer, int8_t value)
    {
        assert(buffer.index + 1 <= buffer.size);
        *((buffer.data + buffer.index)) = (value);
        buffer.index += 1;
    }

    void write_s16(Buffer& buffer, int16_t value)
    {
        assert(buffer.index + 2 <= buffer.size);
        *((int16_t*)(buffer.data + buffer.index)) = host_to_network(value);
        buffer.index += 2;
    }

    void write_s32(Buffer& buffer, int32_t value)
    {
        assert(buffer.index + 4 <= buffer.size);
        *((int32_t*)(buffer.data + buffer.index)) = host_to_network(value);
        buffer.index += 4;
    }

    void write_s64(Buffer& buffer, int64_t value)
    {
        assert(buffer.index + 8 <= buffer.size);
        *((int64_t*)(buffer.data + buffer.index)) = host_to_network(value);
        buffer.index += 8;
    }

    int8_t read_s8(Buffer& buffer)
    {
        assert(buffer.index + 1 <= buffer.size);
        int8_t value;
        value = (*((buffer.data + buffer.index)));
        buffer.index += 1;
        return value;
    }
    int16_t read_s16(Buffer& buffer)
    {
        assert(buffer.index + 2 <= buffer.size);
        int16_t value;
        value = network_to_host(*((int16_t*)(buffer.data + buffer.index)));
        buffer.index += 2;
        return value;
    }
    int32_t read_s32(Buffer& buffer)
    {
        assert(buffer.index + 4 <= buffer.size);
        int32_t value;
        value = network_to_host(*((int32_t*)(buffer.data + buffer.index)));
        buffer.index += 4;
        return value;
    }

    int64_t read_s64(Buffer& buffer)
    {
        assert(buffer.index + 8 <= buffer.size);
        int64_t value;
        value = network_to_host(*((int64_t*)(buffer.data + buffer.index)));
        buffer.index += 8;
        return value;
    }

    using J3ML::ReinterpretAs;

    float u16_to_float(uint16_t value)
    {
        union {
            float f_number;
            uint16_t uint16_arr[2];
        } union_for_conv;
        union_for_conv.f_number = 0;
        union_for_conv.uint16_arr[1] = value;
        return union_for_conv.f_number;
    }

    uint16_t float_to_u16(float value)
    {
        union {
            float f_number;
            uint16_t uint16_arr[2];
        } union_for_conv;
        union_for_conv.f_number = value;
        uint16_t a = union_for_conv.uint16_arr[1];
        return a;
    }

    void write_f16(Buffer& buffer, float value) {
        write_u16(buffer, float_to_u16(value));
    }

    void write_f32(Buffer &buffer, float value) {
        write_u32(buffer, ReinterpretAs<uint32_t>(value));
    }

    void write_f64(Buffer &buffer, double value) {
        write_u64(buffer, ReinterpretAs<uint64_t>(value));
    }

    float read_f16(Buffer& buffer) {
        return u16_to_float(read_u16(buffer));
    }

    float read_f32(Buffer &buffer) {
        return ReinterpretAs<float>(read_u32(buffer));
    }

    double read_f64(Buffer &buffer) {
        return ReinterpretAs<float>(read_u64(buffer));
    }

    void write_string(Buffer &buffer, std::string value) {
        uint16_t length = value.length();
        write_u16(buffer, length);
        for (size_t i = 0; i < length; i++)
        {
            write_u8(buffer, (uint8_t)value[i]);
        }
    }

    std::string read_string(Buffer &buffer) {
        uint16_t str_size = read_u16(buffer);

        std::string result;
        for (size_t i = 0; i < str_size; i++)
        {
            result += (char)read_u8(buffer);
        }

        return result;
    }

    ByteBuffer &ByteBuffer::operator<<(uint8_t &val) {
        write_u8(buff, val);
        return *this;
    }

    ByteBuffer &ByteBuffer::operator<<(uint16_t &val) {
        write_u16(buff, val);
        return *this;
    }

    ByteBuffer &ByteBuffer::operator<<(uint32_t &val) {
        write_u32(buff, val);
        return *this;
    }
}